# -*- encoding: utf-8 -*-
#########################################################################
#This module intergrates Open ERP with the magento core                 #
#Core settings are stored here                                          #
#########################################################################
#                                                                       #
# Copyright (C) 2009  Sharoon Thomas, Raphaël Valyi                     #
# Copyright (C) 2011 Akretion Sébastien BEAU sebastien.beau@akretion.com#
# Copyright (C) 2011-2012 Camptocamp Guewen Baconnier                   #
#                                                                       #
#This program is free software: you can redistribute it and/or modify   #
#it under the terms of the GNU General Public License as published by   #
#the Free Software Foundation, either version 3 of the License, or      #
#(at your option) any later version.                                    #
#                                                                       #
#This program is distributed in the hope that it will be useful,        #
#but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#GNU General Public License for more details.                           #
#                                                                       #
#You should have received a copy of the GNU General Public License      #
#along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
#########################################################################

import string
import copy

from osv import osv, fields
from tools.translate import _
import magerp_osv

class res_partner_category(magerp_osv.magerp_osv):
    _inherit = "res.partner.category"

    _columns = {
                    'tax_class_id':fields.integer('Tax Class ID'),
                }
res_partner_category()

class res_partner_address(magerp_osv.magerp_osv):
    _inherit = "res.partner.address"

    _columns = {
                    'firstname':fields.char('First Name', size=100),
                    'lastname':fields.char('Last Name', size=100),
                    'is_magento_order_address':
                        fields.boolean('Magento Order Address',
                                       help="Created for a Magento sale order",
                                       readonly=True),
                    'magento_address_book':
                        fields.boolean('Magento Address Book',
                                       help="Part of the Magento Address Book",
                                       readonly=True)
                }

    _defaults = {
        'is_magento_order_address': lambda * a: False,
        }

    def mage_import_base(self,cr,uid,conn, external_referential_id, defaults=None, context=None):
        # quick hack
        # remove the id so the address will be imported using "list method"
        # instead of "get method" because the entity_id field miss in the get method.
        context.pop('id', False)
        return super(res_partner_address, self).mage_import_base(
            cr, uid, conn, external_referential_id, defaults=defaults, context=context)

res_partner_address()

class res_partner(magerp_osv.magerp_osv):
    _inherit = "res.partner"

    def _is_magento_exported(self, cr, uid, ids, field_name, arg, context=None):
        """Return True if the partner is already exported to at least one magento shop
        """
        res = {}
        # get all magento external_referentials
        referentials = self.pool.get('external.referential').search(cr, uid, [('magento_referential', '=', True)])
        for partner in self.browse(cr, uid, ids, context):
            for referential in referentials:
                res[partner.id] = False
                if self.oeid_to_extid(cr, uid, partner.id, referential, context):
                    res[partner.id] = True
                    break
        return res

    _columns = {
                    'group_id':fields.many2one('res.partner.category', 'Magento Group(Category)'),
                    'store_id':fields.many2one('magerp.storeviews', 'Last Store View', readonly=True, help="Last store view where the customer has bought."),
                    'store_ids':fields.many2many('magerp.storeviews', 'magerp_storeid_rel', 'partner_id', 'store_id', 'Store Views', readonly=True),
                    'website_id':fields.many2one('external.shop.group', 'Magento Website', help='Select a website for which the Magento customer will be bound.'),
                    'created_in':fields.char('Created in', size=100),
                    'created_at':fields.datetime('Created Date'),
                    'updated_at':fields.datetime('Updated At'),
                    'emailid':fields.char('Email Address', size=100, help="Magento uses this email ID to match the customer. If filled, if a Magento customer is imported from the selected website with the exact same email, he will be bound with this partner and this latter will be updated with Magento's values."),
                    'mag_vat':fields.char('Magento VAT', size=50, help="To be able to receive customer VAT number you must set it in Magento Admin Panel, menu System / Configuration / Client Configuration / Name and Address Options."),
                    'mag_birthday':fields.date('Birthday', help="To be able to receive customer birthday you must set it in Magento Admin Panel, menu System / Configuration / Client Configuration / Name and Address Options."),
                    'mag_newsletter':fields.boolean('Newsletter'),
                    'magento_exported': fields.function(_is_magento_exported, type="boolean", method=True, string="Exists on Magento"),
                    'is_magento_guest': fields.boolean(
                        'Is a Magento guest',
                        help="This partner has been created for a Magento "
                             "guest sale order.",
                        readonly=True)
                }

    _sql_constraints = [('emailid_uniq', 'unique(emailid, website_id)', 'A partner already exists with this email address on the selected website.')]

    def _import_magento_addresses(self, cr, uid, connection, ext_customer_id,
                                  referential_id, context=None):
        """
        Import all the magento addresses of a magento customer's address book

        :param connection: connection to Magento
        :param int ext_customer_id: id on Magento of the customer
        :param int referential_id: id of the Magento external referential
        """
        if context is None:
            context = {}
        address_obj = self.pool.get('res.partner.address')
        import_ctx = context.copy()
        import_ctx['ids_or_filter'] = [ext_customer_id]
        defaults = {'magento_address_book': True}
        return address_obj.get_external_data(
            cr, uid, connection, referential_id,
            defaults=defaults, context=import_ctx)

    def _disable_missing_addresses(self, cr, uid, partner_id, curr_address_ids,
                                   referential_id, context=None):
        """
        Disable the addresses of the partner which have been deleted on
        Magento since last synchronization.

        :param int partner_id: id of the partner for whom we
            update the addresses
        :param list curr_address_ids: list of openerp ids of the addresses
            which currently exist in the magento address book
        :param int referential_id: id of the Magento external referential
        :return: True
        """
        # list of the all the addresses in openerp which have been imported
        # from this referential
        address_obj = self.pool.get('res.partner.address')
        partner_address_ids = address_obj.search(
            cr, uid, [('partner_id', '=', partner_id)], context=context)
        all_ref_address_ids = \
            [address_id for address_id in partner_address_ids
             if address_obj.oeid_to_extid(
                cr, uid, address_id, referential_id, context=context)]

        # do a diff to know which addresses does not anymore exist
        # on Magento
        del_address_ids = set(all_ref_address_ids).difference(curr_address_ids)
        address_obj.write(
            cr, uid, del_address_ids, {'active': False}, context=context)
        return True

    def import_magento_address_book(self, cr, uid, ids, referential_id,
                                    context=None):
        """
        Import all the addresses from the Magento referential's address book
        Deactivate the addresses which have been deleted from Magento
        since the last import.

        :param int/list ids: ids or list of ids of partner to import the
            address book
        :param int referential_id: id of the Magento external referential
        """
        if context is None:
            context = {}

        if isinstance(ids, (int, long)):
            ids = [ids]

        ref_obj = self.pool.get('external.referential')
        connection = context.get('conn_obj') or \
                        ref_obj.external_connection(
                            cr, uid, referential_id, context=context)

        for partner_id in ids:
            ext_customer_id = self.oeid_to_extid(
                cr, uid, partner_id, referential_id, context=context)

            result = self._import_magento_addresses(
                cr, uid, connection, ext_customer_id, referential_id,
                context=context)

            # list of all addresses currently in the magento address book
            curr_address_ids = result['create_ids'] + result['write_ids']
            self._disable_missing_addresses(
                cr, uid, partner_id, curr_address_ids, referential_id,
                context=context)

        return True

    def add_magento_vat_number(self, cr, uid, partner_id, tax_vat, country_code=None, context=None):
        """Adds vat number (country code+magento vat) if base_vat module is installed and Magento sends customer_taxvat
        TODO replace me by a generic function maybe the best solution will to have a field vat and a flag vat_ok and a flag force vat_ok
        And so it's will be possible to have an invalid vat number (imported from magento for exemple) but the flag will be not set
        Also we should think about the way to update customer. Indeed by default there are never updated"""
        partner_vals = {'mag_vat': tax_vat}
        cr.execute('select * from ir_module_module where name=%s and state=%s', ('base_vat','installed'))
        if cr.fetchone():
            allchars = string.maketrans('', '')
            delchars = ''.join([c for c in allchars if c not in string.letters + string.digits])
            vat = tax_vat.translate(allchars, delchars).upper()
            vat_country, vat_number = vat[:2].lower(), vat[2:]
            vat_ok = False
            if 'check_vat_' + vat_country in dir(self):
                check = getattr(self, 'check_vat_' + vat_country)
                vat_ok = check(vat_number)
            else:
                # Maybe magento vat number has not country code prefix. Take it from billing address.
                if country_code:
                    fnct = 'check_vat_' + country_code.lower()
                    if fnct in dir(self):
                        check = getattr(self, fnct)
                        vat_ok = check(vat)
                        vat = country_code + vat
                    else:
                        vat_ok = False
            if vat_ok:
                partner_vals.update({'vat_subjected':True, 'vat':vat})
        self.write(cr, uid, [partner_id], partner_vals)
        return True

    def write(self, cr, uid, ids, vals, context=None):
        if isinstance(ids, (int, long)):
            ids = [ids]
        if vals.get('store_id') and not vals.get('store_ids'):
            for partner_id in ids:
                # Add the last store id visited if
                # it is not in the list of store ids
                # partner store_id = last store where he bought (so this order)
                # partner store_ids = all stores where he bought
                atomic_vals = copy.deepcopy(vals)
                partner = self.browse(cr, uid, partner_id, context=context)
                if partner.store_id:
                    last_store_id = partner.store_id.id
                    if last_store_id not in [st.id for st in partner.store_ids]:
                        atomic_vals['store_ids'] = [(4, last_store_id)]

                super(res_partner, self).write(
                    cr, uid, partner_id, atomic_vals, context=context)
        else:
            return super(res_partner, self).write(
                cr, uid, ids, vals, context=context)
        return True

    def create(self, cr, uid, vals, context=None):
        if vals.get('store_id') and not vals.get('store_ids'):
            vals['store_ids'] = [(4, vals['store_id'])]
        return super(res_partner, self).create(cr, uid, vals, context=context)

    def search_magento_partner(self, cr, uid, magento_email, website_id, is_bound=None, context=None):
        """
        Searches and returns the partners which are potentially the same
        based on the mail and the magento website id

        :param str magento_email: mail of the customer on magento
        :param int website_id: id of the website (openerp id!)
        :param bool is_bound: specify it the partner must already be bound
            with a magento customer (True) or not (False)
            When None value: all partners ids are returned
        :return: id of the found partner or False if any partner's found
        """
        partner_ids = False
        if website_id:
            partner_ids = self.search(cr, uid,
                                      [('emailid', '=', magento_email),
                                       ('website_id', '=', website_id)],
                                      context=context)

        # when the partner has been created, the website is not necessarily
        # filled, in such case we search the first occurrence
        # with the same email not already bound with a magento customer
        if not partner_ids:
            partner_ids = self.search(cr, uid,
                                      [('emailid', '=', magento_email),
                                       ('website_id', '=', False)],
                                      context=context)
        # keep or exclude already bound partners according to argument is_bound
        if is_bound is not None:
            partner_ids = [partner.id for partner
                           in self.browse(cr, uid, partner_ids[:], context=context)
                           if bool(partner.magento_exported) is is_bound]

        return partner_ids and partner_ids[0] or False

    def _existing_oeid_for_extid_import(self, cr, uid, vals, external_id, external_referential_id, context=None):
        """
        When importing a new customer from Magento (no external_id found in ir.model.data), we search
        for a partner with the same emailid and website than the imported one
        (fields must be filled in advance).

        The external customer will update the partner returned in the tuple.
        The external id will be created on this partner with the external resource
        if the value is False in the tuple.

        :param dict vals: vals to create in OpenERP, already evaluated by oevals_from_extdata
        :param int external_id: external id of the resource to create
        :param int external_referential_id: external referential id from where we import the resource
        :return: tuple of (ir.model.data id / False: external id to create, model resource id / False: resource to create)
        """
        existing_ir_model_data_id, existing_res_id = super(res_partner, self)\
        ._existing_oeid_for_extid_import(cr, uid, vals, external_id, external_referential_id, context=context)

        if existing_res_id:
            return existing_ir_model_data_id, existing_res_id

        partner_id = self.search_magento_partner(
            cr, uid,
            vals['emailid'],
            vals.get('website_id', False),
            is_bound=None,
            context=context)

        return False, partner_id

res_partner()
